import argparse

from avi.protobuf.common_pb2 import AviObjectType
from avi.protobuf.metrics_query_pb2 import MetricsQuery
from avi.protobuf.options_pb2 import IpAddr
from avi.protobuf.options_pb2 import IpAddrType
from avi.protobuf.options_pb2 import MetricsEntityType
from avi.protobuf.pool_pb2 import LbAlgorithm
from avi.protobuf.pool_pb2 import Pool
from avi.protobuf.vs_pb2 import VirtualService
from avi.sdk.avi_sdk import AviSdk


class VirtualserviceSDKExample(object):
    def create_basic_vs(self, vs_name, vip, servers=None):
        # create pool protobuf object
        pool_pb = Pool()
        pool_pb.name = vs_name + 'pool_pb'
        pool_pb.lb_algorithm = LbAlgorithm.Value('LB_ALGORITHM_LEAST_CONNECTIONS')

        server_protos = []
        for server in servers:
            parts = server.split(':')
            ip_addr = parts[0]
            print(parts[1])
            port = parts[1] if len(parts) == 2 else 80
            server_proto = pool_pb.servers.add()
            ip = IpAddr()
            ip.addr = ip_addr
            ip.type = IpAddrType.Value('V4')
            server_proto.ip.CopyFrom(ip)
            print(port)
            server_proto.port = int(port)

        # post the pool protobuf with pool resource
        pool_resource = sdk.resource(AviObjectType.POOL)
        pool_obj = pool_resource.post(pool_pb)

        # create vs protobuf
        vs_pb = VirtualService()
        vs_pb.name = vs_name
        service = vs_pb.services.add()
        service.port = 80
        ip = IpAddr()
        ip.addr = vip
        ip.type = IpAddrType.Value('V4')
        vs_pb.ip_address.CopyFrom(ip)
        vs_pb.pool_uuid = pool_obj.protobuf().uuid

        # post vs protobuf using vs resource
        vs_obj = vs_resource.post(vs_pb)

        print('Virtual service created successfully')


    def scaleout_vs(self, vs_name):
        # get the vs avi object by name
        vs_obj = vs_resource.get_by_name(vs_name)
        # avi object will have the actions for the entities hence scaleout can be called on vs_obj
        new_vs_obj = vs_obj.scaleout()
        print('Virtual service scaled out successfully')

    def delete_vs(self, name):
        vs_resource.delete(name=name)
        print('Virtual service deleted successfully')

    def get_metrics(self, metric_entity, metric_entity_uuid, metric_ids):
        # To get the metric data need to create metric query object
        mq_req = MetricsQuery()  # protobuf
        mq_req.metric_entity = metric_entity
        mq_req.entity_uuid = metric_entity_uuid
        mq_req.metric_id = metric_ids
        mq_req.step = 300
        mq_req.limit = 12

        # metric data can be fetched from analytics resource
        analytics_resource = sdk.resource(AviObjectType.ANALYTICS)
        metric_obj = analytics_resource.get(mq_req)
        print(metric_obj)

    def get_vs_inventory(self, name):
        vs_obj = vs_resource.get_by_name()
        # inventory for every entity is available in the avi object of that entity
        return vs_obj.inventory()
        print(inventory)



if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('-o', '--option',
                        choices=['create-basic-vs', 'show-vs-inventory',  'delete-vs',
                                 'show-vs-metric', 'scaleout-vs'],
                        help='lastest timestamp',
                        default='create-basic-vs')
    parser.add_argument('-n', '--vs_name_suffix',
                        help='VirtualService Name Suffix',
                        default='vs')
    parser.add_argument(
        '-s', '--server_ips',
        help='Pool Server IPs comma separated Eg. 1.1.1.1,2.2.2.2',
        default='1.1.1.1,1.1.1.2')
    parser.add_argument('-u', '--user', help='controller user',
                        default='admin')
    parser.add_argument('-p', '--password', help='controller user password',
                        default='avi123')
    parser.add_argument('-t', '--tenant', help='tenant name',
                        default='admin')
    parser.add_argument('-c', '--controller_ip', help='controller ip')
    parser.add_argument('-i', '--vip', help='VIP address')

    parser.add_argument('-r', '--resource_name',
                        help='Resource Name to be deleted',
                        default='basic-vs')
    parser.add_argument('-m', '--metric_id',
                        help='Comma separated metric ids',
                        default='l4_client.avg_bandwidth')


    args = parser.parse_args()
    sdk = AviSdk('10.10.24.152', 'admin', 'avi123')
    vs_resource = sdk.resource(AviObjectType.VIRTUALSERVICE)
    vse = VirtualserviceSDKExample()
    servers = [server.strip() for server in args.server_ips.split(',')]
    if args.option == 'create-basic-vs':
        vip = args.vip if args.vip else '10.10.42.2'
        vs_name = 'basic-' + args.vs_name_suffix
        vse.create_basic_vs(vs_name, vip, servers)
    elif args.option == 'show-vs-metric':
        vs_obj = vs_resource.get_by_name(args.resource_name)
        vse.get_metric(MetricsEntityType.VSERVER_METRICS_ENTITY, vs_obj.protobuf().uuid , args.metric_id)
    elif args.option == 'delete-vs':
        vse.delete(args.resource_name)
    elif args.option == 'show-vs-inventory':
        inventory = vse.get_vs_inventory(args.resource_name)





#!/usr/bin/env python
from avi.sdk.avi_api import ApiSession
from avi.sdk.avi_api import APIError
import getpass
import argparse
import sys
import json
import pprint
import time
'''
Overview:

Accelerated networking has been introduced as an SE group level option in 17.2.14 release. By default all SE groups created 17.2.14 onwards will have this option enabled, but this option cannot be enabled on an existing SE group having SEs created with accelerated nw disabled. This is because 
Azure does not allow mixing VMs with accelerated nw enabled and disabled in the same availability set.


Migration to accelerated networking 

Create a new SE group with Accelerated networking enabled. All the existing VSes need to be migrated to this new SE group. Follow one or a combination o the following strategies to migrate the Virtual services to the new SE group.
1. Create a clone of existing Virtual services in the new SE group. Once the new VSes are up, switch the DNS record for each VS to point to the new Virtual services. This strategy makes sure this is no downtime for the clients. 
2. Modify the SE group for the Virtual services. This strategy needs the Virtual services to be disabled before changing the SE group, causing downtime for the clients, but allows all the analytical history of the Virtual services to be preserved.
A combination of both the strategies above can be followed to gain the benefits of both of them.

Please use the script provided to clone the virtual services to a new SE group.

Avi cloud connector does periodic reconcile of the existing virtual services to the cloud, this reconcile should be disabled while cloning the virtual services to a new se group. To disable cloud_reconcile, execute the following command on controller shell.

>configure controller properties
>no cloud_reconcile
>save

Setup a new SE group in the cloud where the Virtual services need to be cloned.

Execute the script provided as follows to create clones of virtual services.

usage: clone_vs_segroup.py [-h] [-c CONTROLLER] [-u USER] [-p PASSWORD]
                           [-x API_VERSION] [-C CLOUD] [-t TENANT] [-e]
                           [-f FROMSEGROUP] [-d TOSEGROUP] [-s SUFFIX] [-n]

optional arguments:
  -h, --help            show this help message and exit
  -c CONTROLLER, --controller CONTROLLER
                        FQDN or IP address of Avi Vantage controller
  -u USER, --user USER  Avi Vantage username
  -p PASSWORD, --password PASSWORD
                        Avi Vantage password
  -x API_VERSION, --api_version API_VERSION
                        Avi Vantage API version (default=17.2.14)
  -C CLOUD, --cloud CLOUD
                        Cloud name
  -t TENANT, --tenant TENANT
                        Tenant name
  -e, --enable          Enable the cloned Virtual Service
  -f FROMSEGROUP, --fromsegroup FROMSEGROUP
                        Source SE group
  -d TOSEGROUP, --tosegroup TOSEGROUP
                        Destination SE group
  -s SUFFIX, --suffix SUFFIX
                        suffix to be added to virtual service names
  -n, --dryrun          Dry run mode

The script will clone the existing virtual services to new SE group created. Once the VSes are up, the older VSes can be deleted or disabled.
'''

import urllib3
urllib3.disable_warnings()
print('\n\n\n\n')
pp = pprint.PrettyPrinter()


dry_run_mode = False
tenant = 'admin'
DEFAULT_API_VERSION = '17.2.14'


def get_ref_url_by_name(controller_ip, name, objtype, tenant='admin'):
    return 'https://%s/api/%s?name=%s&tenant=%s' % (controller_ip, objtype, name, tenant)


def switchto_cloud(api, cloud):
    api.headers['X-Avi-Cloud'] = cloud
    api.cloud = cloud


def get_all_vs(api):
    'To Get all vs with paginiation'
    all_vs = []
    api_path = "virtualservice?fields=name'"
    while True:
        vs_data = api.get(api_path).json()
        if 'results' in vs_data:
            all_vs.extend(vs_data['results'])
        if 'next' in vs_data:
            api_path = vs_data['next'].split('/')[-1]
            print("Next page : %s" % api_path)
            continue
        else:
            break
    return all_vs


def main(args):

    if args:

        # If not specified on the command-line, prompt the user for the
        # controller IP address and/or password

        controller_ip = args.controller
        user = args.user
        password = args.password

        while not controller_ip:
            controller_ip = input('Controller:')

        while not password:
            password = getpass.getpass('Password for %s@%s:' %
                                       (user, controller_ip))
    api = ApiSession.get_session(
        controller_ip, user, password, tenant=args.tenant, api_version=args.api_version)
    cloud = args.cloud
    from_segrp = args.fromsegroup
    to_segrp = args.tosegroup
    switchto_cloud(api, cloud)
    dry_run_mode = args.dryrun
    suffix = args.suffix
    enabled = args.enable

    all_vs = get_all_vs(api)
    print(all_vs)
    old_se_grp_uuid = api.get('serviceenginegroup?name=%s' % from_segrp).json()[
        'results'][0]['uuid']
    print('se_grp_uuid : %s' % old_se_grp_uuid)
    new_se_grp = api.get('serviceenginegroup?name=%s' %
                         to_segrp).json()['results'][0]
    print('new_se_grp_uuid : %s' % new_se_grp['uuid'])
    c_vs_vips = {}
    all_vsvips = {}
    all_nw_policies = []
    all_vses = []
    all_http_pols = []
    for vs in all_vs:
        shared_vip = False
        vs_uuid = vs['uuid']
        vs_api = 'virtualservice/%s?join_subresources=runtime&page_size=1000' % vs_uuid
        vs = api.get(vs_api).json()
        vs_segid = vs['se_group_ref'].split('/')[-1]
        if vs_segid != old_se_grp_uuid:
            #print('Skipping VS as it does not belong to se group %s: %s' %
            #      (args.fromsegroup, vs['name']))
            continue
        print('Cloning VS %s to segrp %s' % (vs['name'], args.tosegroup))
        vs.pop('runtime')
        vs.pop('vip_runtime')

        #if not vs.has_key('vsvip_ref'):
        if not ('vsvip_ref' in vs):
            continue

        old_vsvip_ref = vs['vsvip_ref']
        #if c_vs_vips.has_key(old_vsvip_ref):
        if old_vsvip_ref in c_vs_vips:
            shared_vip = True
        else:
            c_vs_vips[old_vsvip_ref] = True
        vsvip = api.get('vsvip/%s' % old_vsvip_ref.split('/')[-1]).json()
        vsvip['vip'][0].pop('ip_address')
        if 'floating_ip' in vsvip['vip'][0]:
        #if vsvip['vip'][0].has_key('floating_ip'):
            vsvip['vip'][0].pop('floating_ip')
        vsvip['name'] = vsvip['name'] + suffix
        vsvip.pop('uuid')
        if 'use_standard_alb' in new_se_grp:
            vsvip['use_standard_alb'] = new_se_grp['use_standard_alb']
        vsvip_post = None
        if not dry_run_mode:
            if shared_vip:
                vsvip_post = api.get_object_by_name('vsvip', vsvip['name'])
            else:
                try:
                    vsvip_post = api.post('vsvip', vsvip).json()
                except APIError as e:
                    if e.rsp.status_code == 409:
                        vsvip_post = api.get_object_by_name('vsvip', vsvip['name'])
                    else: 
                        raise e
                except Exception as e:
                    raise e
            vsvip_ref = api.get_obj_ref(vsvip_post)
            print(vsvip_post)
        else:
            vsvip_ref = get_ref_url_by_name(
                controller_ip, vsvip['name'], 'vsvip', tenant=tenant)
        all_vsvips[vsvip_ref] = vsvip

        #if vs.has_key('http_policies'):
        if 'http_policies' in vs:
            http_policies = vs['http_policies']
            if http_policies:
                for pol in http_policies:
                    pol_obj = api.get(
                        'httppolicyset/%s' % pol['http_policy_set_ref'].split('/')[-1]).json()
                    pol_obj['name'] = pol_obj['name'] + suffix
                    pol_obj.pop('uuid')
                    if not dry_run_mode:
                        try:
                            new_pol_obj = api.post('httppolicyset', pol_obj)
                        except APIError as e:
                                if e.rsp.status_code == 409:
                                    new_pol_obj = api.get_object_by_name('httppolicyset', pol_obj['name'])
                                    pass
                                else: 
                                    raise e
                        http_pol_ref = api.get_obj_ref(new_pol_obj)
                    else:
                        http_pol_ref = get_ref_url_by_name(
                            controller_ip, pol_obj['name'], 'httppolicyset')
                    all_http_pols.append(pol_obj)
                    pol['http_policy_set_ref'] = http_pol_ref
        #if vs.has_key('network_security_policy_ref'):
        if 'network_security_policy_ref' in vs:
            netw_sec_pol_ref = vs['network_security_policy_ref']
            netw_sec_pol = api.get('networksecuritypolicy/%s' %
                                   netw_sec_pol_ref.split('/')[-1]).json()
            netw_sec_pol.pop('uuid')
            netw_sec_pol['name'] = netw_sec_pol['name'] + suffix
            netw_sec_pol_post = None
            if not dry_run_mode:
                try:
                    netw_sec_pol_post = api.post(
                        'networksecuritypolicy', netw_sec_pol).json()
                except APIError as e:
                    if e.rsp.status_code == 409:
                        netw_sec_pol_post = api.get_object_by_name('networksecuritypolicy', netw_sec_pol['name'])
                        pass
                    else:
                        raise e
                nw_pol_ref = api.get_obj_ref(netw_sec_pol_post)
            else:
                nw_pol_ref = get_ref_url_by_name(
                    controller_ip, netw_sec_pol['name'], 'networksecuritypolicy')
            all_nw_policies.append(netw_sec_pol)
        else:
            nw_pol_ref = None

        vs['name'] = vs['name'] + suffix
        vs['vsvip_ref'] = vsvip_ref
        vs['enabled'] = enabled
        if not dry_run_mode:
            vs['vip'][0]['ip_address']['addr'] = vsvip_post['vip'][0]['ip_address']['addr']
        vs['network_security_policy_ref'] = nw_pol_ref
        vs['se_group_ref'] = api.get_obj_ref(new_se_grp)
        vs.pop('uuid')
        if not dry_run_mode:
            try:
                time.sleep(0.5)
                vs_post = api.post('virtualservice', vs).json()
            except APIError as e:
                if e.rsp.status_code == 409:
                    pass
                else: 
                    raise e
        all_vses.append(vs)
    for vs in all_vs:
        vs_uuid = vs['uuid']
        vs_api = 'virtualservice/%s?join_subresources=runtime&page_size=1000' % vs_uuid
        vs = api.get(vs_api).json()
        vs.pop('runtime')
        vs.pop('vip_runtime')

        #if vs.has_key('vsvip_ref'):
        if 'vsvip_ref' in vs:
            #print('Skipping VS %s as it has its own VIP' % vs['name'])
            continue
        if vs['type'] != 'VS_TYPE_VH_CHILD':
            #print('%s is not child vs' % vs['name'])
            continue
        parent_vs = api.get('virtualservice/%s' %
                            vs['vh_parent_vs_ref'].split('/')[-1]).json()
        vs_segid = parent_vs['se_group_ref'].split('/')[-1]
        if vs_segid != old_se_grp_uuid:
            #print('Skipping VS %s as it does not belong to se group %s' %
            #      (vs['name'], args.fromsegroup))
            continue
        print('checking vs %s' % vs['name'])
        new_parent = parent_vs['name'] + suffix
        vs['name'] = vs['name'] + suffix
        vs.pop('uuid')
        vs['enabled'] = enabled
        #if vs.has_key('http_policies'):
        if 'http_policies' in vs:
            http_policies = vs['http_policies']
            if http_policies:
                for pol in http_policies:
                    pol_obj = api.get(
                        'httppolicyset/%s' % pol['http_policy_set_ref'].split('/')[-1]).json()
                    pol_obj['name'] = pol_obj['name'] + suffix
                    pol_obj.pop('uuid')
                    if not dry_run_mode:
                        try:
                            new_pol_obj = api.post('httppolicyset', pol_obj)
                        except APIError as e:
                            if e.rsp.status_code == 409:
                                new_pol_obj = api.get_object_by_name('httppolicyset', pol_obj['name'])
                                pass
                            else: 
                                raise e
                        http_pol_ref = api.get_obj_ref(new_pol_obj)
                    else:
                        http_pol_ref = get_ref_url_by_name(
                            controller_ip, pol_obj['name'], 'httppolicyset')
                    all_http_pols.append(pol_obj)
                    pol['http_policy_set_ref'] = http_pol_ref
        #if vs.has_key('network_security_policy_ref'):
        if 'network_security_policy_ref' in vs:
            netw_sec_pol_ref = vs['network_security_policy_ref']
            netw_sec_pol = api.get('networksecuritypolicy/%s' %
                                   netw_sec_pol_ref.split('/')[-1]).json()
            netw_sec_pol.pop('uuid')
            netw_sec_pol['name'] = netw_sec_pol['name'] + suffix
            netw_sec_pol_post = None
            if not dry_run_mode:
                try:
                    netw_sec_pol_post = api.post(
                        'networksecuritypolicy', netw_sec_pol).json()
                except APIError as e:
                        if e.rsp.status_code == 409:
                            netw_sec_pol_post = api.get_object_by_name('networksecuritypolicy', netw_sec_pol['name'])
                            pass
                        else: 
                            raise e
                nw_pol_ref = api.get_obj_ref(netw_sec_pol_post)
            else:
                nw_pol_ref = get_ref_url_by_name(
                    controller_ip, netw_sec_pol['name'], 'networksecuritypolicy')
            vs['network_security_policy_ref'] = nw_pol_ref
            all_nw_policies.append(netw_sec_pol)
        else:
            nw_pol_ref = None

        if not dry_run_mode:
            new_parent_vs = api.get('virtualservice?name=%s' % new_parent).json()[
                'results'][0]
            parent_vs_ref = api.get_obj_ref(new_parent_vs)
            vs['vh_parent_vs_ref'] = parent_vs_ref
            try:
                vs_post = api.post('virtualservice', vs)
            except APIError as e:
                    if e.rsp.status_code == 409:
                        pass
                    else: 
                        raise e
            print('posted vs %s' % vs_post)
        else:
            parent_vs_ref = get_ref_url_by_name(
                controller_ip, new_parent, 'virtualservice')
            vs['vh_parent_vs_ref'] = parent_vs_ref
        all_vses.append(vs)

    with open('vsvip.json', mode='w') as outfile:
#        json.dump(all_vsvips.values(), outfile, indent=4, sort_keys=True)
        json.dump(all_vsvips, outfile, indent=4, sort_keys=True)
    with open('nwpolicies.json', mode='w') as outfile:
        json.dump(all_nw_policies, outfile, indent=4, sort_keys=True)
    with open('httppolicies.json', mode='w') as outfile:
        json.dump(all_http_pols, outfile, indent=4, sort_keys=True)
    with open('vs.json', mode='w') as outfile:
        json.dump(all_vses, outfile, indent=4, sort_keys=True)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='')
    parser.add_argument('-c', '--controller',
                        help='FQDN or IP address of Avi Vantage controller')
    parser.add_argument('-u', '--user', help='Avi Vantage username',
                        default='admin')
    parser.add_argument('-p', '--password', help='Avi Vantage password')
    parser.add_argument('-x', '--api_version',
                        help='Avi Vantage API version (default=%s)' % DEFAULT_API_VERSION,
                        default=DEFAULT_API_VERSION)
    parser.add_argument('-C', '--cloud',
                        help='Cloud name', default='Default-Cloud')
    parser.add_argument('-t', '--tenant',
                        help='Tenant name', default='admin')
    parser.add_argument('-e', '--enable',
                        help='Enable the cloned Virtual Service', action='store_true')
    parser.add_argument('-f', '--fromsegroup', help='Source SE group')
    parser.add_argument('-d', '--tosegroup', help='Destination SE group')
    parser.add_argument(
        '-s', '--suffix', help='suffix to be added to virtual service names')
    parser.add_argument(
        '-n', '--dryrun', help='Dry run mode', action='store_true')
    args = parser.parse_args()
    if not args.fromsegroup or not args.tosegroup:
        parser.print_help(sys.stderr)
        sys.exit(1)

    main(args)



import os
import sys
import copy
import json
import logging
import time
import urlparse
from datetime import datetime, timedelta
from requests import ConnectionError
from requests import Response
from requests.exceptions import ChunkedEncodingError
from requests.sessions import Session
from ssl import SSLError


logger = logging.getLogger(__name__)

global sessionDict
sessionDict = {}


def avi_timedelta(td):
    """
    This is a wrapper class to workaround python 2.6 builtin datetime.timedelta
    does not have total_seconds method
    :param td timedelta object
    """
    if type(td) != timedelta:
        raise TypeError()
    if sys.version_info >= (2, 7):
        ts = td.total_seconds()
    else:
        ts = td.seconds + (24 * 3600 * td.days)
    return ts


def avi_sdk_syslog_logger(logger_name='avi.sdk'):
    # The following sets up syslog module to log underlying avi SDK messages
    # based on the environment variables:
    #   AVI_LOG_HANDLER: names the logging handler to use. Only syslog is
    #     supported.
    #   AVI_LOG_LEVEL: Logging level used for the avi SDK. Default is DEBUG
    #   AVI_SYSLOG_ADDRESS: Destination address for the syslog handler.
    #   Default is /dev/log
    from logging.handlers import SysLogHandler
    lf = '[%(asctime)s] %(levelname)s [' \
         '%(module)s.%(funcName)s:%(lineno)d] %(message)s'
    log = logging.getLogger(logger_name)
    log_level = os.environ.get('AVI_LOG_LEVEL', 'DEBUG')
    if log_level:
        log.setLevel(getattr(logging, log_level))
    formatter = logging.Formatter(lf)
    sh = SysLogHandler(address=os.environ.get('AVI_SYSLOG_ADDRESS',
                                              '/dev/log'))
    sh.setFormatter(formatter)
    log.addHandler(sh)
    return log


class ObjectNotFound(Exception):
    pass


class APIError(Exception):
    def __init__(self, arg, rsp=None):
        self.args = [arg, rsp]
        self.rsp = rsp


class AviServerError(APIError):
    def __init__(self, arg, rsp=None):
        super(AviServerError, self).__init__(arg, rsp)


class APINotImplemented(Exception):
    pass


class ApiResponse(Response):
    """
    Returns copy of the requests.Response object provides additional helper
    routines
        1. obj: returns dictionary of Avi Object
    """
    def __init__(self, rsp):
        super(ApiResponse, self).__init__()
        for k, v in list(rsp.__dict__.items()):
            setattr(self, k, v)

    def json(self):
        """
        Extends the session default json interface to handle special errors
        and raise Exceptions
        returns the Avi object as a dictionary from rsp.text
        """
        if self.status_code in (200, 201):
            if not self.text:
                # In cases like status_code == 201 the response text could be
                # empty string.
                return None
            return super(ApiResponse, self).json()
        elif self.status_code == 204:
            # No response needed; e.g., delete operation
            return None
        elif self.status_code == 404:
            raise ObjectNotFound('HTTP Error: %d Error Msg %s' % (
                self.status_code, self.text), self)
        elif self.status_code >= 500:
            raise AviServerError('HTTP Error: %d Error Msg %s' % (
                self.status_code, self.text), self)
        else:
            raise APIError('HTTP Error: %d Error Msg %s' % (
                self.status_code, self.text), self)

    def count(self):
        """
        return the number of objects in the collection response. If it is not
        a collection response then it would simply return 1.
        """
        obj = self.json()
        if 'count' in obj:
            # this was a resposne to collection
            return obj['count']
        return 1

    @staticmethod
    def to_avi_response(resp):
        if type(resp) == Response:
            return ApiResponse(resp)
        return resp


class AviCredentials(object):
    controller = ''
    username = ''
    password = ''
    api_version = '16.4.4'
    tenant = None
    tenant_uuid = None
    token = None
    port = None
    timeout = 300
    session_id = None
    csrftoken = None

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def update_from_ansible_module(self, m):
        """
        :param m: ansible module
        :return:
        """
        if m.params.get('avi_credentials'):
            for k, v in m.params['avi_credentials'].items():
                if hasattr(self, k):
                    setattr(self, k, v)
        if m.params['controller']:
            self.controller = m.params['controller']
        if m.params['username']:
            self.username = m.params['username']
        if m.params['password']:
            self.password = m.params['password']
        if (m.params['api_version'] and
                (m.params['api_version'] != '16.4.4')):
            self.api_version = m.params['api_version']
        if m.params['tenant']:
            self.tenant = m.params['tenant']
        if m.params['tenant_uuid']:
            self.tenant_uuid = m.params['tenant_uuid']
        if m.params.get('session_id'):
            self.session_id = m.params['session_id']
        if m.params.get('csrftoken'):
            self.csrftoken = m.params['csrftoken']

    def __str__(self):
        return 'controller %s user %s api %s tenant %s' % (
            self.controller, self.username, self.api_version, self.tenant)


class ApiSession(Session):
    """
    Extends the Request library's session object to provide helper
    utilities to work with Avi Controller like authentication, api massaging
    etc.
    """

    # This keeps track of the process which created the cache.
    # At anytime the pid of the process changes then it would create
    # a new cache for that process.
    AVI_SLUG = 'Slug'
    SESSION_CACHE_EXPIRY = 20*60
    SHARED_USER_HDRS = ['X-CSRFToken', 'Session-Id', 'Referer', 'Content-Type']
    MAX_API_RETRIES = 3

    def __init__(self, controller_ip=None, username=None, password=None,
                 token=None, tenant=None, tenant_uuid=None, verify=False,
                 port=None, timeout=60, api_version=None,
                 retry_conxn_errors=True, data_log=False,
                 avi_credentials=None, session_id=None, csrftoken=None,
                 lazy_authentication=False, max_api_retries=None):
        """
         ApiSession takes ownership of avi_credentials and may update the
         information inside it.

        Initialize new session object with authenticated token from login api.
        It also keeps a cache of user sessions that are cleaned up if inactive
        for more than 20 mins.

        Notes:
        01. If mode is https and port is none or 443, we don't embed the
            port in the prefix. The prefix would be 'https://ip'. If port
            is a non-default value then we concatenate https://ip:port
            in the prefix.
        02. If mode is http and the port is none or 80, we don't embed the
            port in the prefix. The prefix would be 'http://ip'. If port is
            a non-default value, then we concatenate http://ip:port in
            the prefix.
        """
        super(ApiSession, self).__init__()
        logger.debug("Creating session with following values:\n "
                     "controller_ip: %s, username: %s, tenant: %s, "
                     "tenant_uuid: %s, verify: %s, port: %s, timeout: %s, "
                     "api_version: %s, retry_conxn_errors: %s, data_log: %s,"
                     "avi_credentials: %s, session_id: %s, csrftoken: %s,"
                     "lazy_authentication: %s, max_api_retries: %s"
                     % (controller_ip, username, tenant,
                        tenant_uuid, verify, port,
                        timeout, api_version, retry_conxn_errors,
                        data_log, avi_credentials, session_id,
                        csrftoken, lazy_authentication, max_api_retries))
        if not avi_credentials:
            tenant = tenant if tenant else "admin"
            self.avi_credentials = AviCredentials(
                controller=controller_ip, username=username,
                password=password, api_version=api_version,
                tenant=tenant, tenant_uuid=tenant_uuid,
                token=token, port=port, timeout=timeout,
                session_id=session_id, csrftoken=csrftoken)
        else:
            self.avi_credentials = avi_credentials
        self.headers = {}
        self.verify = verify
        self.retry_conxn_errors = retry_conxn_errors
        self.remote_api_version = {}
        self.user_hdrs = {}
        self.data_log = data_log
        self.num_session_retries = 0
        self.num_api_retries = 0
        self.retry_wait_time = 0
        self.max_session_retries = (
            self.MAX_API_RETRIES if max_api_retries is None
            else int(max_api_retries))
        # Refer Notes 01 and 02
        k_port = port if port else 443
        if self.avi_credentials.controller.startswith('http'):
            k_port = 80 if not self.avi_credentials.port else k_port
            if self.avi_credentials.port is None or \
                    self.avi_credentials.port == 80:
                self.prefix = self.avi_credentials.controller
            else:
                self.prefix = '{x}:{y}'.format(
                    x=self.avi_credentials.controller,
                    y=self.avi_credentials.port)
        else:
            if port is None or port == 443:
                self.prefix = 'https://{x}'.format(
                    x=self.avi_credentials.controller)
            else:
                self.prefix = 'https://{x}:{y}'.format(
                    x=self.avi_credentials.controller,
                    y=self.avi_credentials.port)
        self.timeout = timeout
        self.key = '%s:%s:%s' % (self.avi_credentials.controller,
                                 self.avi_credentials.username, k_port)
        # Added api token and session id to sessionDict for handle single
        # session
        if self.avi_credentials.csrftoken:
            sessionDict[self.key] = {
                'api': self,
                "csrftoken": self.avi_credentials.csrftoken,
                "session_id": self.avi_credentials.session_id,
                "last_used": datetime.utcnow()
            }
        elif lazy_authentication:
            sessionDict.get(self.key, {}).update(
                {'api': self, "last_used": datetime.utcnow()})
        else:
            self.authenticate_session()
        self.num_session_retries = 0
        self.pid = os.getpid()
        ApiSession._clean_inactive_sessions()
        return

    @property
    def controller_ip(self):
        return self.avi_credentials.controller

    @controller_ip.setter
    def controller_ip(self, controller_ip):
        self.avi_credentials.controller = controller_ip

    @property
    def username(self):
        return self.avi_credentials.username

    @property
    def connected(self):
        return sessionDict.get(self.key, {}).get('connected', False)

    @username.setter
    def username(self, username):
        self.avi_credentials.username = username

    @property
    def password(self):
        return self.avi_credentials.password

    @password.setter
    def password(self, password):
        self.avi_credentials.password = password

    @property
    def keystone_token(self):
        return sessionDict.get(self.key, {}).get('csrftoken', None)

    @keystone_token.setter
    def keystone_token(self, token):
        sessionDict[self.key]['csrftoken'] = token

    @property
    def tenant_uuid(self):
        self.avi_credentials.tenant_uuid

    @tenant_uuid.setter
    def tenant_uuid(self, tenant_uuid):
        self.avi_credentials.tenant_uuid = tenant_uuid

    @property
    def tenant(self):
        return self.avi_credentials.tenant

    @tenant.setter
    def tenant(self, tenant):
        if tenant:
            self.avi_credentials.tenant = tenant
        else:
            self.avi_credentials.tenant = 'admin'

    @property
    def port(self):
        self.avi_credentials.port

    @port.setter
    def port(self, port):
        self.avi_credentials.port = port

    @property
    def api_version(self):
        return self.avi_credentials.api_version

    @api_version.setter
    def api_version(self, api_version):
        self.avi_credentials.api_version = api_version

    @property
    def session_id(self):
        return sessionDict[self.key]['session_id']

    def get_context(self):
        return {
            'session_id': sessionDict[self.key]['session_id'],
            'csrftoken': sessionDict[self.key]['csrftoken']
        }

    @staticmethod
    def clear_cached_sessions():
        global sessionDict
        sessionDict = {}

    @staticmethod
    def get_session(
            controller_ip=None, username=None, password=None, token=None,
            tenant=None, tenant_uuid=None, verify=False, port=None, timeout=60,
            retry_conxn_errors=True, api_version=None, data_log=False,
            avi_credentials=None, session_id=None, csrftoken=None,
            lazy_authentication=False, max_api_retries=None, idp_class=None):
        """
        returns the session object for same user and tenant
        calls init if session dose not exist and adds it to session cache
        :param controller_ip: controller IP address
        :param username:
        :param password:
        :param token: Token to use; example, a valid keystone token
        :param tenant: Name of the tenant on Avi Controller
        :param tenant_uuid: Don't specify tenant when using tenant_id
        :param port: Rest-API may use a different port other than 443
        :param timeout: timeout for API calls; Default value is 60 seconds
        :param retry_conxn_errors: retry on connection errors
        :param api_version: Controller API version
        :param idp_class: IDP class. Currently supports OKtaSAMLApiSession,
        OneloginApiSession
        """

        if not idp_class:
            idp_class = ApiSession
        else:
            if not ("ApiSession" in str(idp_class.__base__)):
                raise APIError("idp_class {} not valid class. Please provide "
                               "correct idp class. Base class of idp class is "
                               "{}".format(idp_class, str(idp_class.__base__)))
        # Validate input idp_class
        if not avi_credentials:
            tenant = tenant if tenant else "admin"
            avi_credentials = AviCredentials(
                controller=controller_ip, username=username,
                password=password, api_version=api_version,
                tenant=tenant, tenant_uuid=tenant_uuid,
                token=token, port=port, timeout=timeout,
                session_id=session_id, csrftoken=csrftoken)

        k_port = avi_credentials.port if avi_credentials.port else 443
        if avi_credentials.controller.startswith('http'):
            k_port = 80 if not avi_credentials.port else k_port
        key = '%s:%s:%s' % (avi_credentials.controller,
                            avi_credentials.username, k_port)
        cached_session = sessionDict.get(key)
        if cached_session:
            user_session = cached_session['api']
            if not (user_session.avi_credentials.csrftoken or
                    lazy_authentication):
                user_session.authenticate_session()
        else:
            user_session = idp_class(
                controller_ip, username, password, token=token,
                tenant=tenant, tenant_uuid=tenant_uuid,
                verify=verify, port=port, timeout=timeout,
                retry_conxn_errors=retry_conxn_errors,
                api_version=api_version, data_log=data_log,
                avi_credentials=avi_credentials,
                lazy_authentication=lazy_authentication,
                max_api_retries=max_api_retries)
        return user_session

    def reset_session(self):
        """
        resets and re-authenticates the current session.
        """
        sessionDict[self.key]['connected'] = False
        logger.info('resetting session for %s', self.key)
        self.user_hdrs = {}
        for k, v in self.headers.items():
            if k not in self.SHARED_USER_HDRS:
                self.user_hdrs[k] = v
        self.headers = self.user_hdrs
        self.authenticate_session()

    def authenticate_session(self):
        """
        Performs session authentication with Avi controller and stores
        session cookies and sets header options like tenant.
        """
        body = {"username": self.avi_credentials.username}
        if self.avi_credentials.password:
            body["password"] = self.avi_credentials.password
        elif self.avi_credentials.token:
            body["token"] = self.avi_credentials.token
        else:
            raise APIError("Neither user password or token provided")
        logger.debug('authenticating user %s prefix %s',
                     self.avi_credentials.username, self.prefix)
        self.cookies.clear()
        err = None
        try:
            rsp = super(ApiSession, self).post(
                self.prefix+"/login", body, timeout=self.timeout,
                verify=self.verify)

            if rsp.status_code == 200:
                self.num_session_retries = 0
                self.remote_api_version = rsp.json().get('version', {})
                session_cookie_name = rsp.json().get(
                    'session_cookie_name', 'sessionid')
                self.headers.update(self.user_hdrs)
                if rsp.cookies and 'csrftoken' in rsp.cookies:
                    csrftoken = rsp.cookies['csrftoken']
                    sessionDict[self.key] = {
                        'csrftoken': csrftoken,
                        'session_id': rsp.cookies[session_cookie_name],
                        'last_used': datetime.utcnow(),
                        'api': self,
                        'connected': True
                    }
                    self.avi_credentials.csrftoken = csrftoken
                    self.avi_credentials.session_id = rsp.cookies[
                        session_cookie_name]
                logger.debug("authentication success for user %s",
                             self.avi_credentials.username)
                return
            # Check for bad request and invalid credentials response code
            elif rsp.status_code in [401, 403]:
                logger.error('Status Code %s msg %s' % (
                    rsp.status_code, rsp.text))
                err = APIError('Status Code %s msg %s' % (
                    rsp.status_code, rsp.text), rsp)
                raise err
            else:
                logger.error("Error status code %s msg %s", rsp.status_code,
                             rsp.text)
                err = APIError('Status Code %s msg %s' % (
                    rsp.status_code, rsp.text), rsp)
                raise err
        except (ConnectionError, SSLError, ChunkedEncodingError) as e:
            if not self.retry_conxn_errors:
                raise
            logger.warning('Connection error retrying %s', e)
            err = e
        # comes here only if there was either exception or login was not
        # successful
        if self.retry_wait_time:
            time.sleep(self.retry_wait_time)
        self.num_session_retries += 1
        if self.num_session_retries > self.max_session_retries:
            self.num_session_retries = 0
            logger.error("giving up after %d retries connection failure %s" % (
                self.max_session_retries, True))
            raise err
        self.authenticate_session()
        return

    def _get_api_headers(self, tenant, tenant_uuid, timeout, headers,
                         api_version):
        """
        returns the headers that are passed to the requests.Session api calls.
        """
        api_hdrs = copy.deepcopy(self.headers)
        api_hdrs.update({
            "Referer": self.prefix,
            "Content-Type": "application/json"
        })
        api_hdrs['timeout'] = str(timeout)
        if self.key in sessionDict and 'csrftoken' in \
                sessionDict.get(self.key):
            api_hdrs['X-CSRFToken'] = sessionDict.get(self.key)['csrftoken']
        else:
            self.authenticate_session()
            api_hdrs['X-CSRFToken'] = sessionDict.get(self.key)['csrftoken']
        if api_version:
            api_hdrs['X-Avi-Version'] = api_version
        elif self.avi_credentials.api_version:
            api_hdrs['X-Avi-Version'] = self.avi_credentials.api_version
        if tenant:
            tenant_uuid = None
        elif tenant_uuid:
            tenant = None
        else:
            tenant = self.avi_credentials.tenant
            tenant_uuid = self.avi_credentials.tenant_uuid
        if tenant_uuid:
            api_hdrs.update({"X-Avi-Tenant-UUID": "%s" % tenant_uuid})
            api_hdrs.pop("X-Avi-Tenant", None)
        elif tenant:
            api_hdrs.update({"X-Avi-Tenant": "%s" % tenant})
            api_hdrs.pop("X-Avi-Tenant-UUID", None)
        # Override any user headers that were passed by users. We don't know
        # when the user had updated the user_hdrs
        if self.user_hdrs:
            api_hdrs.update(self.user_hdrs)
        if headers:
            # overwrite the headers passed via the API calls.
            api_hdrs.update(headers)
        return api_hdrs

    def _api(self, api_name, path, tenant, tenant_uuid, data=None,
             headers=None, timeout=None, api_version=None, **kwargs):
        """
        It calls the requests.Session APIs and handles session expiry
        and other situations where session needs to be reset.
        returns ApiResponse object
        :param path: takes relative path to the AVI api.
        :param tenant: overrides the tenant used during session creation
        :param tenant_uuid: overrides the tenant or tenant_uuid during session
            creation
        :param timeout: timeout for API calls; Default value is 60 seconds
        :param headers: dictionary of headers that override the session
            headers.
        """
        if self.pid != os.getpid():
            logger.info('pid %d change detected new %d. Closing session',
                        self.pid, os.getpid())
            self.close()
            self.pid = os.getpid()
        if timeout is None:
            timeout = self.timeout
        fullpath = self._get_api_path(path)
        fn = getattr(super(ApiSession, self), api_name)
        api_hdrs = self._get_api_headers(tenant, tenant_uuid, timeout, headers,
                                         api_version)
        connection_error = False
        err = None
        cookies = {
            'csrftoken': api_hdrs['X-CSRFToken'],
        }
        try:
            sessionid = sessionDict[self.key]['session_id']
            cookies['sessionid'] = sessionid
            cookies['avi-sessionid'] = sessionid
        except KeyError:
            pass
        try:
            if (data is not None) and (type(data) == dict):
                resp = fn(fullpath, data=json.dumps(data), headers=api_hdrs,
                          timeout=timeout, cookies=cookies, **kwargs)
            else:
                resp = fn(fullpath, data=data, headers=api_hdrs,
                          timeout=timeout, cookies=cookies, **kwargs)
        except (ConnectionError, SSLError, ChunkedEncodingError) as e:
            logger.warning('Connection error retrying %s', e)
            if not self.retry_conxn_errors:
                raise
            connection_error = True
            err = e
        except Exception as e:
            logger.error('Error in Requests library %s', e)
            raise
        if not connection_error:
            logger.debug('path: %s http_method: %s hdrs: %s params: '
                         '%s data: %s rsp: %s', fullpath, api_name.upper(),
                         api_hdrs, kwargs, data,
                         (resp.text if self.data_log else 'None'))
        if connection_error or resp.status_code in (401, 419):
            if connection_error:
                try:
                    self.close()
                except Exception as e:
                    # ignoring exception in cleanup path
                    pass
                logger.warning('Connection failed, retrying.')
                # Adding sleep before retrying
                if self.retry_wait_time:
                    time.sleep(self.retry_wait_time)
            else:
                logger.info('received error %d %s so resetting connection',
                            resp.status_code, resp.text)
            ApiSession.reset_session(self)
            self.num_api_retries += 1
            if self.num_api_retries > self.max_session_retries:
                # Added this such that any code which re-tries can succeed
                # eventually.
                self.num_api_retries = 0
                if not connection_error:
                    err = APIError('Status Code %s msg %s' % (
                        resp.status_code, resp.text), resp)
                logger.error(
                    "giving up after %d retries conn failure %s err %s" % (
                        self.max_session_retries, connection_error, err))
                raise err
            # should restore the updated_hdrs to one passed down
            resp = self._api(api_name, path, tenant, tenant_uuid, data,
                             headers=headers, api_version=api_version,
                             timeout=timeout, **kwargs)
            self.num_api_retries = 0

        if resp.cookies and 'csrftoken' in resp.cookies:
            csrftoken = resp.cookies['csrftoken']
            self.headers.update({"X-CSRFToken": csrftoken})
        self._update_session_last_used()
        return ApiResponse.to_avi_response(resp)

    def get_controller_details(self):
        result = {
            "controller_ip": self.controller_ip,
            "controller_api_version": self.remote_api_version
        }
        return result

    def get(self, path, tenant='', tenant_uuid='', timeout=None, params=None,
            api_version=None, **kwargs):
        """
        It extends the Session Library interface to add AVI API prefixes,
        handle session exceptions related to authentication and update
        the global user session cache.
        :param path: takes relative path to the AVI api.
        :param tenant: overrides the tenant used during session creation
        :param tenant_uuid: overrides the tenant or tenant_uuid during session
            creation
        :param timeout: timeout for API calls; Default value is 60 seconds
        :param params: dictionary of key value pairs to be sent as query
            parameters
        :param api_version: overrides x-avi-header in request header during
            session creation
        get method takes relative path to service and kwargs as per Session
            class get method
        returns session's response object
        """
        return self._api('get', path, tenant, tenant_uuid, timeout=timeout,
                         params=params, api_version=api_version, **kwargs)

    def get_object_by_name(self, path, name, tenant='', tenant_uuid='',
                           timeout=None, params=None, api_version=None,
                           **kwargs):
        """
        Helper function to access Avi REST Objects using object
        type and name. It behaves like python dictionary interface where it
        returns None when the object is not present in the AviController.
        Internally, it transforms the request to api/path?name=<name>...
        :param path: relative path to service
        :param name: name of the object
        :param tenant: overrides the tenant used during session creation
        :param tenant_uuid: overrides the tenant or tenant_uuid during session
            creation
        :param timeout: timeout for API calls; Default value is 60 seconds
        :param params: dictionary of key value pairs to be sent as query
            parameters
        :param api_version: overrides x-avi-header in request header during
            session creation
        returns dictionary object if successful else None
        """
        obj = None
        if not params:
            params = {}
        params['name'] = name
        resp = self.get(path, tenant=tenant, tenant_uuid=tenant_uuid,
                        timeout=timeout,
                        params=params, api_version=api_version, **kwargs)
        if resp.status_code in (401, 419):
            ApiSession.reset_session(self)
            resp = self.get_object_by_name(
                path, name, tenant, tenant_uuid, timeout=timeout,
                params=params, **kwargs)
        if resp.status_code > 499 or 'Invalid version' in resp.text:
            logger.error('Error in get object by name for %s named %s. '
                         'Error: %s' % (path, name, resp.text))
            raise AviServerError(resp.text, rsp=resp)
        elif resp.status_code > 299:
            return obj
        try:
            if 'results' in resp.json():
                obj = resp.json()['results'][0]
            else:
                # For apis returning single object eg. api/cluster
                obj = resp.json()
        except IndexError:
            logger.warning('Warning: Object Not found for %s named %s' %
                           (path, name))
            obj = None
        self._update_session_last_used()
        return obj

    def post(self, path, data=None, tenant='', tenant_uuid='', timeout=None,
             force_uuid=None, params=None, api_version=None, **kwargs):
        """
        It extends the Session Library interface to add AVI API prefixes,
        handle session exceptions related to authentication and update
        the global user session cache.
        :param path: takes relative path to the AVI api.It is modified by
        the library to conform to AVI Controller's REST API interface
        :param data: dictionary of the data. Support for json string
            is deprecated
        :param tenant: overrides the tenant used during session creation
        :param tenant_uuid: overrides the tenant or tenant_uuid during session
            creation
        :param timeout: timeout for API calls; Default value is 60 seconds
        :param params: dictionary of key value pairs to be sent as query
            parameters
        :param api_version: overrides x-avi-header in request header during
            session creation
        returns session's response object
        """
        if force_uuid is not None:
            headers = kwargs.get('headers', {})
            headers[self.AVI_SLUG] = force_uuid
            kwargs['headers'] = headers
        return self._api('post', path, tenant, tenant_uuid, data=data,
                         timeout=timeout, params=params,
                         api_version=api_version, **kwargs)

    def put(self, path, data=None, tenant='', tenant_uuid='',
            timeout=None, params=None, api_version=None, **kwargs):
        """
        It extends the Session Library interface to add AVI API prefixes,
        handle session exceptions related to authentication and update
        the global user session cache.
        :param path: takes relative path to the AVI api.It is modified by
            the library to conform to AVI Controller's REST API interface
        :param data: dictionary of the data. Support for json string
            is deprecated
        :param tenant: overrides the tenant used during session creation
        :param tenant_uuid: overrides the tenant or tenant_uuid during session
            creation
        :param timeout: timeout for API calls; Default value is 60 seconds
        :param params: dictionary of key value pairs to be sent as query
            parameters
        :param api_version: overrides x-avi-header in request header during
            session creation
        returns session's response object
        """
        return self._api('put', path, tenant, tenant_uuid, data=data,
                         timeout=timeout, params=params,
                         api_version=api_version, **kwargs)

    def patch(self, path, data=None, tenant='', tenant_uuid='',
              timeout=None, params=None, api_version=None, **kwargs):
        """
        It extends the Session Library interface to add AVI API prefixes,
        handle session exceptions related to authentication and update
        the global user session cache.
        :param path: takes relative path to the AVI api.It is modified by
            the library to conform to AVI Controller's REST API interface
        :param data: dictionary of the data. Support for json string
            is deprecated
        :param tenant: overrides the tenant used during session creation
        :param tenant_uuid: overrides the tenant or tenant_uuid during session
            creation
        :param timeout: timeout for API calls; Default value is 60 seconds
        :param params: dictionary of key value pairs to be sent as query
            parameters
        :param api_version: overrides x-avi-header in request header during
            session creation
        returns session's response object
        """
        return self._api('patch', path, tenant, tenant_uuid, data=data,
                         timeout=timeout, params=params,
                         api_version=api_version, **kwargs)

    def put_by_name(self, path, name, data=None, tenant='',
                    tenant_uuid='', timeout=None, params=None,
                    api_version=None, **kwargs):
        """
        Helper function to perform HTTP PUT on Avi REST Objects using object
        type and name.
        Internally, it transforms the request to api/path?name=<name>...
        :param path: relative path to service
        :param name: name of the object
        :param data: dictionary of the data. Support for json string
            is deprecated
        :param tenant: overrides the tenant used during session creation
        :param tenant_uuid: overrides the tenant or tenant_uuid during session
            creation
        :param timeout: timeout for API calls; Default value is 60 seconds
        :param params: dictionary of key value pairs to be sent as query
            parameters
        :param api_version: overrides x-avi-header in request header during
            session creation
        returns session's response object
        """
        uuid = self._get_uuid_by_name(
            path, name, tenant, tenant_uuid, api_version=api_version)
        path = '%s/%s' % (path, uuid)
        return self.put(path, data, tenant, tenant_uuid, timeout=timeout,
                        params=params, api_version=api_version, **kwargs)

    def delete(self, path, tenant='', tenant_uuid='',
               timeout=None, params=None, data=None,
               api_version=None, **kwargs):
        """
        It extends the Session Library interface to add AVI API prefixes,
        handle session exceptions related to authentication and update
        the global user session cache.
        :param path: takes relative path to the AVI api.It is modified by
        the library to conform to AVI Controller's REST API interface
        :param tenant: overrides the tenant used during session creation
        :param tenant_uuid: overrides the tenant or tenant_uuid during session
            creation
        :param timeout: timeout for API calls; Default value is 60 seconds
        :param params: dictionary of key value pairs to be sent as query
            parameters
        :param data: dictionary of the data. Support for json string
            is deprecated
        :param api_version: overrides x-avi-header in request header during
            session creation
        returns session's response object
        """
        return self._api('delete', path, tenant, tenant_uuid, data=data,
                         timeout=timeout, params=params,
                         api_version=api_version, **kwargs)

    def delete_by_name(self, path, name, tenant='', tenant_uuid='',
                       timeout=None, params=None, api_version=None, **kwargs):
        """
        Helper function to perform HTTP DELETE on Avi REST Objects using object
        type and name.Internally, it transforms the request to
        api/path?name=<name>...
        :param path: relative path to service
        :param name: name of the object
        :param tenant: overrides the tenant used during session creation
        :param tenant_uuid: overrides the tenant or tenant_uuid during session
            creation
        :param timeout: timeout for API calls; Default value is 60 seconds
        :param params: dictionary of key value pairs to be sent as query
            parameters
        :param api_version: overrides x-avi-header in request header during
            session creation
        returns session's response object
        """
        uuid = self._get_uuid_by_name(path, name, tenant, tenant_uuid,
                                      api_version=api_version)
        if not uuid:
            raise ObjectNotFound("%s/?name=%s" % (path, name))
        path = '%s/%s' % (path, uuid)
        return self.delete(path, tenant, tenant_uuid, timeout=timeout,
                           params=params, api_version=api_version, **kwargs)

    def get_obj_ref(self, obj):
        """returns reference url from dict object"""
        if not obj:
            return None
        if isinstance(obj, Response):
            obj = json.loads(obj.text)
        if obj.get(0, None):
            return obj[0]['url']
        elif obj.get('url', None):
            return obj['url']
        elif obj.get('results', None):
            return obj['results'][0]['url']
        else:
            return None

    def get_obj_uuid(self, obj):
        """returns uuid from dict object"""
        if not obj:
            raise ObjectNotFound('Object %s Not found' % (obj))
        if isinstance(obj, Response):
            obj = json.loads(obj.text)
        if obj.get('uuid', None):
            return obj['uuid'].split('#')[0]
        elif obj.get('results', None):
            return obj['results'][0]['uuid'].split('#')[0]
        else:
            return None

    def _paginator(self, *args, **kwargs):
        """
        This is a private method to simulate an iterator
        on top of the pagination functionality provided by Avi Controller.

        :param *args: Accepts all args accepted by get method of class
            ApiSession.
        :params **kwargs: Accepts all kwargs accepted by get method of class
            ApiSession.

        Yields batch of objects in the range of 0 to page_size passed in kwargs.
        """
        resp = self.get(*args, **kwargs).json()
        if 'results' in resp:
            yield resp['results']
        else:
            # For apis returning single object eg. api/cluster
             yield [resp]

        page = 2 # Initialized to 2 for new page requests
        while resp.get('next', None) is not None:
            kwargs['params']['page'] = page
            resp = self.get(*args, **kwargs).json()
            page += 1
            yield resp['results']

    def get_objects_iter(self, obj_type, tenant='', tenant_uuid='',
                         timeout=None, params=None, api_version=None, **kwargs):
        """
        Iterator to fetch objects of any type from Avi Controller.
        By default, 100 objects are fetched in every batch. However, user can
        override this behaviour by passing page_size as params.

        Irrespective of page_size the method will yield only one object for
        each iteration.

        :param obj_type: Type of object that needs to be fetched from
			Avi Controller. Ex: pool, virtualservice
        :param tenant: overrides the tenant used during session creation.
        :param tenant_uuid: overrides the tenant or tenant_uuid during session
            creation.
        :param timeout: timeout for API calls; Default value is 60 seconds.
        :param params: dictionary of key value pairs to be sent as query
            parameters.
        :param api_version: overrides x-avi-header in request header during
            session creation.
        :param **kwargs: Accepts any other params accepted by Session class get
			method.

        Yields objects of type obj_type one by one. Example

        >>> # Looping over get_objects_iter
        >>> all_pools_iter = get_objects_iter('pool', params={'fields':'name'})
        >>> for pool in all_pools_iter:
        ...  #Do something with pool
        ...  print(pool)
        {u'name': u'pool-avi',
        u'server_count': 0,
        u'url': u'https://192.10.20.30/api/pool/<uuid>',
        u'uuid': u'<uuid>'}
        """
        if params is None:
            params={}

        if params.get('page_size', None) is None:
            params['page_size'] = 100

        page_size = params.get('page_size')
        if int(page_size) > 200:
            raise ValueError('page_size cannot be more than 200')
        pages_iter =  self._paginator(obj_type, tenant=tenant, params=params,
                                      tenant_uuid=tenant_uuid, timeout=timeout,
                                      api_version=api_version, **kwargs)
        for page in pages_iter:
            for obj in page:
                yield obj

    def get_slug_from_uri(self, uri):
        """return uuid/slug from URI"""
        if not uri or '/' not in uri:
            return uri
        parsed = urlparse.urlparse(uri)
        path = parsed.path
        uuid = os.path.basename(path)
        if '#' in uuid:
            uuid = uuid.split('#')[0]
        return uuid

    def _get_api_path(self, path, uuid=None):
        """
        This function returns the full url from relative path and uuid.
        """
        if path == 'logout':
            return self.prefix+'/'+path
        elif uuid:
            return self.prefix+'/api/'+path+'/'+uuid
        else:
            return self.prefix+'/api/'+path

    def _get_uuid_by_name(self, path, name, tenant='admin',
                          tenant_uuid='', api_version=None):
        """gets object by name and service path and returns uuid"""
        resp = self.get_object_by_name(
            path, name, tenant, tenant_uuid, api_version=api_version)
        if not resp:
            raise ObjectNotFound("%s/%s" % (path, name))
        return self.get_obj_uuid(resp)

    def _update_session_last_used(self):
        if self.key in sessionDict:
            sessionDict[self.key]["last_used"] = datetime.utcnow()

    @staticmethod
    def _clean_inactive_sessions():
        """Removes sessions which are inactive more than 20 min"""
        session_cache = sessionDict
        logger.debug("cleaning inactive sessions in pid %d num elem %d",
                     os.getpid(), len(session_cache))
        for key, session in list(session_cache.items()):
            tdiff = avi_timedelta(datetime.utcnow() - session["last_used"])
            if tdiff < ApiSession.SESSION_CACHE_EXPIRY:
                continue
            try:
                session["api"].post("logout")
            except Exception as e:
                logger.warning("Session not found on controller "
                               "for session ID: %s %s",
                               session, e)
                pass
            del session_cache[key]
            logger.debug("Cleaned inactive session : %s", key)

    def delete_session(self):
        """ Removes the session for cleanup"""
        logger.debug("Removed session for : %s", self.key)
        sessionDict.pop(self.key, None)
        return
# End of file



from avi.sdk.avi_api import ApiSession, \
    sessionDict, APIError
import requests
import re
import urllib
import urlparse
import json
from datetime import datetime
from requests import ConnectionError
from requests.exceptions import ChunkedEncodingError
from ssl import SSLError
import time
import logging

logger = logging.getLogger(__name__)


class OneloginSAMLApiSession(ApiSession):
    """
    Extends the ApiSession class to override authentication
    method and provide helper utilities to work with Avi
    Controller and IDPs like onelogin, okta, etc.
    """

    SAML_URL_SUFFIX = "/sso/login"
    # Request RegX
    saml_request_regex = r'<input type=\"hidden\" ' \
                         r'name=\"SAMLRequest\" value=\"(.*?)\"'
    request_relay_state_regex = r'<input type=\"hidden\" ' \
                                r'name=\"RelayState\" value=\"(.*?)\"'
    request_assertion_url_regex = r'<form method=\"post\" action=\"(.*?)\">'
    # Response RegX
    saml_response_regex = r'<input type=\"hidden\" ' \
                          r'name=\"SAMLResponse\" value=\"(.*?)\"'
    response_relay_state_regex = r'<input type=\"hidden\" ' \
                                 r'name=\"RelayState\" value=\"(.*?)\"'
    response_assertion_url_regex = r'<form method=\"post\" ' \
                                   r'action=\"(.*?)\">'

    def __init__(self, controller=None, username=None, password=None,
                 token=None, tenant=None, tenant_uuid=None, verify=False,
                 port=None, timeout=60, api_version=None,
                 retry_conxn_errors=True, data_log=False,
                 avi_credentials=None, session_id=None, csrftoken=None,
                 lazy_authentication=False, max_api_retries=None,
                 idp_cookies=None):
        """
        This extends ApiSession class and overrides authentication method
        for SMAL authentication.

        :param controller: Controller IP
        :param username: IDP username
        :param password: IDP password
        :param token: Controller token
        :param tenant: Overrides the tenant used during session creation
        :param tenant_uuid: Overrides the tenant or tenant_uuid during session
            creation
        :param verify: Boolean flag for SSL verification of url
        :param port: Controller SSO port
        :param timeout: Timeout for API calls; Default value is 60 seconds
        :param api_version: Overrides x-avi-header in request header during
            session creation
        :param retry_conxn_errors: Retry on connection errors
        :param data_log: Data log
        :param avi_credentials: avi_credential object
        :param session_id: Session ID
        :param csrftoken: CSRF Token
        :param lazy_authentication: Lazy_authentication for controller.
        :param max_api_retries: Maximum API retires
        :param idp_cookies: IDP cookies if want to use existing IDP session
        """

        self.idp_cookies = idp_cookies
        super(OneloginSAMLApiSession, self).__init__(
            controller, username, password, token,
            tenant, tenant_uuid, verify,
            port, timeout, api_version,
            retry_conxn_errors, data_log,
            avi_credentials, session_id, csrftoken,
            lazy_authentication, max_api_retries)
        return

    def saml_assertion(self, username, password):
        """
        Perform SAML request from controller to IDPs.
        Establish session with controller and IDP.
        Assert SAML request into the reqeust.
        Get the controller session and IDP session.

        :param username: IDP Username
        :param password: IDP Password
        :return: controller session and IDP response
        """

        # Getting controller session
        controller_session = requests.Session()
        controller_session.verify = False
        saml_controller_url = self.prefix + self.SAML_URL_SUFFIX
        logger.info("Getting SAML request from url: %s", saml_controller_url)
        resp = controller_session.get(saml_controller_url,
                                      allow_redirects=True)
        if resp.status_code != 200:
            logger.error('Status Code %s msg %s' % (
                resp.status_code, resp.text))
            raise APIError('Status Code %s msg %s' % (
                resp.status_code, resp.text), resp)
        # Getting IDP session
        idp_session = requests.Session()
        saml_request_match = re.search(OneloginSAMLApiSession.saml_request_regex, resp.text,
                                       re.M | re.S)
        if not saml_request_match:
            logger.error("SAML request not generated by controller.")
            raise APIError("SAML request not generated by controller.")
        saml_request = saml_request_match.group(1)
        relay_state = re.search(OneloginSAMLApiSession.request_relay_state_regex, resp.text,
                                re.M | re.S).group(1)
        assertion_url = re.search(OneloginSAMLApiSession.request_assertion_url_regex, resp.text,
                                  re.M | re.S).group(1)
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        saml_data = urllib.urlencode({
            'SAMLRequest': saml_request,
            'RelayState': relay_state})
        if self.idp_cookies:
            logger.info("Controller url %s generated SAML request is being "
                        "sent to IDP with existing IDP cookies.",
                        saml_controller_url)
            idp_resp = idp_session.post(assertion_url, headers=headers,
                                        data=saml_data, allow_redirects=False,
                                        cookies=self.idp_cookies)
        else:
            logger.info("Controller url %s generated SAML request is being "
                        "sent to IDP.", saml_controller_url)
            idp_resp = idp_session.post(assertion_url, headers=headers,
                                        data=saml_data,
                                        allow_redirects=False)
        if resp.status_code not in (200, 301, 302):
            logger.error('Status Code %s msg %s' % (
                resp.status_code, resp.text))
            raise APIError('Status Code %s msg %s' % (
                resp.status_code, resp.text), resp)
        if "SAMLResponse" not in idp_resp.text:
            redirect_url = idp_resp.headers['Location']
            idp_resp = idp_session.get(redirect_url,
                                       allow_redirects=False)
            if resp.status_code not in (200, 301, 302):
                logger.error('Status Code %s msg %s' % (
                    resp.status_code, resp.text))
                raise APIError('Status Code %s msg %s' % (
                    resp.status_code, resp.text), resp)
            query_string = idp_resp.headers['Location'].split('=')[1]
            data = {"return": query_string}
            json_data = json.dumps(data)
            headers = {'content-type': 'application/json'}
            parsed_uri = urlparse.urlparse(assertion_url)
            # This needs to be modified for other IDPs.
            auth_url = "{}://{}/access/auth".format(parsed_uri.scheme,
                                                    parsed_uri.netloc)
            resp = idp_session.post(auth_url, headers=headers,
                                    data=json_data)
            if resp.status_code in [401, 403]:
                logger.error('Status Code %s msg Invalid SAML credentials %s'
                             % (resp.status_code, resp.text))
                raise APIError('Status Code %s msg Invalid SAML credentials %s'
                               % (resp.status_code, resp.text), resp)
            elif resp.status_code != 200:
                logger.error('Status Code %s msg %s' % (
                    resp.status_code, resp.text))
                raise APIError('Status Code %s msg %s' % (
                    resp.status_code, resp.text), resp)
            # credentials payload for given IDP
            credentials_tuple = [('username', 'login',
                                  username),
                                 ('password', 'password',
                                  password)]
            for state in credentials_tuple:
                bearer = "Bearer " + resp.text.split('jwt":"')[1][:-3]
                headers = {'content-type': 'application/json',
                           'authorization': bearer}
                user_data = {'state': state[0],
                             'payload': {state[1]: state[2]}}
                json_data = json.dumps(user_data)
                resp = idp_session.put(auth_url, headers=headers,
                                       data=json_data)
                if resp.status_code in [401, 403]:
                    logger.error('Status Code %s msg Invalid SAML credentials %s'
                                 % (resp.status_code, resp.text))
                    raise APIError('Status Code %s msg Invalid SAML credentials %s'
                                   % (resp.status_code, resp.text), resp)
                elif resp.status_code != 200:
                    logger.error('Status Code %s msg %s' % (
                        resp.status_code, resp.text))
                    raise APIError('Status Code %s msg %s' % (
                        resp.status_code, resp.text), resp)
            data = json.loads(resp.text)
            try:
                token = data["request"]["params"]["saml_request_params_token"]
            except KeyError:
                raise APIError("Couldn't complete "
                               "authentication with IDP")
            url = data["request"]["uri"]
            params = {'saml_request_params_token': token}
            resp = idp_session.get(url, params=params)
            if resp.status_code != 200:
                logger.error('Status Code %s msg %s' % (
                    resp.status_code, resp.text))
                raise APIError('Status Code %s msg %s' % (
                    resp.status_code, resp.text), resp)
        return controller_session, resp

    def authenticate_session(self):
        """
        Performs SAML authentication with Avi controller and IDPs.
        Stores session cookies and sets header parameters.
        """

        username = self.avi_credentials.username
        if self.avi_credentials.password:
            password = self.avi_credentials.password
        else:
            raise APIError("No user password provided")
        logger.debug('authenticating user %s prefix %s',
                     self.avi_credentials.username, self.prefix)
        self.cookies.clear()
        try:
            # Assert SAML response
            controller_session, resp = self.saml_assertion(username, password)
            content = resp.text
            saml_response_match = re.search(OneloginSAMLApiSession.saml_response_regex, content,
                                            re.M | re.S)
            saml_response = saml_response_match.group(1)
            relay_state = re.search(OneloginSAMLApiSession.response_relay_state_regex, content, re.M |
                                    re.S).group(1)
            assertion_url = re.search(OneloginSAMLApiSession.response_assertion_url_regex, content, re.M |
                                      re.S).group(1)
            saml_data = urllib.urlencode([
                ('SAMLResponse', saml_response),
                ('RelayState', relay_state)])
            headers = {'Content-Type': 'application/x-www-form-urlencoded'}
            rsp = controller_session.post(assertion_url,
                                          headers=headers,
                                          data=saml_data)
            if rsp.status_code == 200:
                self.num_session_retries = 0
                self.remote_api_version = \
                    rsp.headers.get('AVI_API_VERSION', {})
                self.headers.update(self.user_hdrs)
                if rsp.cookies and 'csrftoken' in rsp.cookies:
                    sessionDict[self.key] = {
                        'csrftoken': rsp.cookies['csrftoken'],
                        'session_id': rsp.cookies['sessionid'],
                        'last_used': datetime.utcnow(),
                        'api': self,
                        'connected': True
                    }
                logger.debug("authentication success for user %s",
                             self.avi_credentials.username)
                return
            # Check for bad request and invalid credentials response code
            elif rsp.status_code in [401, 403]:
                logger.error('Status Code %s msg %s' % (
                    rsp.status_code, rsp.text))
                err = APIError('Status Code %s msg %s' % (
                    rsp.status_code, rsp.text), rsp)
            else:
                logger.error("Error status code %s msg %s", rsp.status_code,
                             rsp.text)
                err = APIError('Status Code %s msg %s' % (
                    rsp.status_code, rsp.text), rsp)
        except (ConnectionError, SSLError, ChunkedEncodingError) as e:
            if not self.retry_conxn_errors:
                raise
            logger.warning('Connection error retrying %s', e)
            err = e
        # comes here only if there was either exception or login was not
        # successful
        if self.retry_wait_time:
            time.sleep(self.retry_wait_time)
        self.num_session_retries += 1
        if self.num_session_retries > self.max_session_retries:
            self.num_session_retries = 0
            logger.error("Giving up after %d retries connection failure %s" % (
                self.max_session_retries, True))
            raise err
        self.authenticate_session()
        return


class OktaSAMLApiSession(ApiSession):
    """
    Extends the ApiSession session class to provide helper
    utilities to work with Avi Controller and IDP for SAML assertion and
    authentication, api massaging, etc
    """

    SAML_URL_SUFFIX = "/sso/login"
    # Request RegX
    saml_request_regex = r'<input type=\"hidden\" ' \
                         r'name=\"SAMLRequest\" value=\"(.*?)\"'
    request_relay_state_regex = r'<input type=\"hidden\" ' \
                                r'name=\"RelayState\" value=\"(.*?)\"'
    request_assertion_url_regex = r'<form method=\"post\" action=\"(.*?)\">'
    # Response RegX
    saml_response_regex = r'<input name=\"SAMLResponse\" ' \
                          r'type=\"hidden\" value=\"(.*?)\"'
    response_relay_state_regex = r'<input name=\"RelayState\" ' \
                                 r'type=\"hidden\" value=\"(.*?)\"'
    response_assertion_url_regex = r'<form id=\"appForm\" ' \
                                   r'action=\"(.*?)\" method=\"post\">'

    def __init__(self, controller=None, username=None, password=None,
                 token=None, tenant=None, tenant_uuid=None, verify=False,
                 port=None, timeout=60, api_version=None,
                 retry_conxn_errors=True, data_log=False,
                 avi_credentials=None, session_id=None, csrftoken=None,
                 lazy_authentication=False, max_api_retries=None,
                 idp_cookies=None):
        """
        This extends ApiSession class and overrides authentication method
        for SMAL authentication.

        :param controller: Controller IP
        :param username: IDP username
        :param password: IDP password
        :param token: Controller token
        :param tenant: Overrides the tenant used during session creation
        :param tenant_uuid: Overrides the tenant or tenant_uuid during session
            creation
        :param verify: Boolean flag for SSL verification of url
        :param port: Controller SSO port
        :param timeout: Timeout for API calls; Default value is 60 seconds
        :param api_version: Overrides x-avi-header in request header during
            session creation
        :param retry_conxn_errors: Retry on connection errors
        :param data_log: Data log
        :param avi_credentials: avi_credential object
        :param session_id: Session ID
        :param csrftoken: CSRF Token
        :param lazy_authentication: Lazy_authentication for controller.
        :param max_api_retries: Maximum API retires
        :param idp_cookies: IDP cookies if want to use existing IDP session
        """

        self.idp_cookies = idp_cookies
        super(OktaSAMLApiSession, self).__init__(
            controller, username, password, token,
            tenant, tenant_uuid, verify,
            port, timeout, api_version,
            retry_conxn_errors, data_log,
            avi_credentials, session_id, csrftoken,
            lazy_authentication, max_api_retries)
        return

    def saml_assertion(self, username, password):
        """
        Perform SAML request from controller to IDPs.
        Establish session with controller and IDP.
        Assert SAML request into the reqeust.
        Get the controller session and IDP session.

        :param username: IDP Username
        :param password: IDP Password
        :return: controller session and IDP response
        """

        # Getting controller session
        controller_session = requests.Session()
        controller_session.verify = False
        saml_controller_url = self.prefix + self.SAML_URL_SUFFIX
        logger.info("Getting SAML request from url: %s", saml_controller_url)
        resp = controller_session.get(saml_controller_url,
                                      allow_redirects=True)
        if resp.status_code != 200:
            logger.error('Status Code %s msg %s' % (
                resp.status_code, resp.text))
            raise APIError('Status Code %s msg %s' % (
                resp.status_code, resp.text), resp)
        saml_request_match = re.search(OktaSAMLApiSession.saml_request_regex, resp.text,
                                       re.M | re.S)
        if not saml_request_match:
            logger.error("SAML request not generated by controller.")
            raise APIError("SAML request not generated by controller.")
        saml_request = saml_request_match.group(1)
        relay_state = re.search(OktaSAMLApiSession.request_relay_state_regex, resp.text,
                                re.M | re.S).group(1)
        assertion_url = re.search(OktaSAMLApiSession.request_assertion_url_regex, resp.text,
                                  re.M | re.S).group(1)

        idp_session = requests.Session()
        idp_session.verify = False
        saml_data = urllib.urlencode({
            'SAMLRequest': saml_request,
            'RelayState': relay_state})
        parsed_uri = urlparse.urlparse(assertion_url)
        base_url = "{}://{}".format(parsed_uri.scheme, parsed_uri.netloc)
        if self.idp_cookies:
            logger.info("Controller url %s generated SAML request is being "
                        "sent to IDP with existing IDP cookies.",
                        saml_controller_url)
            resp = idp_session.get(assertion_url, allow_redirects=False,
                                   cookies=self.idp_cookies)
        else:
            logger.info("Controller url %s generated SAML request is being "
                        "sent to IDP.", saml_controller_url)
            resp = idp_session.get(assertion_url, allow_redirects=False)
        if resp.status_code not in (200, 301, 302):
            logger.error('Status Code %s msg %s' % (
                resp.status_code, resp.text))
            raise APIError('Status Code %s msg %s' % (
                resp.status_code, resp.text), resp)
        if "SAMLResponse" not in resp.text:
            user_data = {"username": username,
                         "options": {"warnBeforePasswordExpired": True,
                                     "multiOptionalFactorEnroll": True},
                         "password": password}
            json_data = json.dumps(user_data)
            headers = {'content-type': 'application/json'}
            resp = idp_session.post(base_url + "/api/v1/authn",
                                    headers=headers,
                                    data=json_data)
            if resp.status_code in [401, 403]:
                logger.error('Status Code %s msg Invalid SAML credentials %s'
                             % (resp.status_code, resp.text))
                raise APIError('Status Code %s msg Invalid SAML credentials %s'
                               % (resp.status_code, resp.text), resp)
            elif resp.status_code != 200:
                logger.error('Status Code %s msg %s' % (
                    resp.status_code, resp.text))
                raise APIError('Status Code %s msg %s' % (
                    resp.status_code, resp.text), resp)
            data = json.loads(resp.text)
            try:
                token = data["sessionToken"]
            except KeyError:
                raise APIError("Couldn't complete authentication with IDP")
            new_url = base_url + "/login/sessionCookieRedirect"
            redirect_url = "{}?{}".format(assertion_url, saml_data)
            params = {'checkAccountSetupComplete': 'true',
                      'token': token,
                      'redirectUrl': redirect_url}
            resp = idp_session.get(new_url, params=params,
                                   allow_redirects=True)
            if resp.status_code not in (200, 301, 302):
                logger.error('Status Code %s msg %s' % (
                    resp.status_code, resp.text))
                raise APIError('Status Code %s msg %s' % (
                    resp.status_code, resp.text), resp)
        return controller_session, resp

    def authenticate_session(self):
        """
        Performs SAML authentication with Avi controller and IDPs.
        Stores session cookies and sets header parameters.
        """

        username = self.avi_credentials.username
        if self.avi_credentials.password:
            password = self.avi_credentials.password
        else:
            raise APIError("No user password provided")
        logger.debug('authenticating user %s prefix %s',
                     self.avi_credentials.username, self.prefix)
        self.cookies.clear()
        try:
            # Assert SAML response
            controller_session, resp = self.saml_assertion(username, password)
            content = resp.text
            saml_response_match = re.search(OktaSAMLApiSession.saml_response_regex,
                                            content, re.M | re.S)
            saml_response = saml_response_match.group(1)
            assertion_url = re.search(OktaSAMLApiSession.response_assertion_url_regex, content,
                                      re.M | re.S |
                                      re.IGNORECASE).group(1)
            relay_state = re.search(OktaSAMLApiSession.response_relay_state_regex, content,
                                    re.M | re.S).group(1)
            from HTMLParser import HTMLParser
            parser = HTMLParser()
            assertion_url = parser.unescape(assertion_url)
            saml_response = parser.unescape(saml_response)
            saml_data = urllib.urlencode([
                ('SAMLResponse', saml_response),
                ('RelayState', relay_state)])
            headers = {'Content-Type': 'application/x-www-form-urlencoded'}
            rsp = controller_session.post(assertion_url,
                                          headers=headers,
                                          data=saml_data,
                                          allow_redirects=True)
            if rsp.status_code == 200:
                self.num_session_retries = 0
                self.remote_api_version = \
                    rsp.headers.get('AVI_API_VERSION', {})
                self.headers.update(self.user_hdrs)
                if rsp.cookies and 'csrftoken' in rsp.cookies:
                    sessionDict[self.key] = {
                        'csrftoken': rsp.cookies['csrftoken'],
                        'session_id': rsp.cookies['sessionid'],
                        'last_used': datetime.utcnow(),
                        'api': self,
                        'connected': True
                    }
                logger.debug("authentication success for user %s",
                             self.avi_credentials.username)
                return
            # Check for bad request and invalid credentials response code
            elif rsp.status_code in [401, 403]:
                logger.error('Status Code %s msg %s' % (
                    rsp.status_code, rsp.text))
                err = APIError('Status Code %s msg %s' % (
                    rsp.status_code, rsp.text), rsp)
                raise err
            else:
                logger.error("Error status code %s msg %s", rsp.status_code,
                             rsp.text)
                err = APIError('Status Code %s msg %s' % (
                    rsp.status_code, rsp.text), rsp)
        except (ConnectionError, SSLError, ChunkedEncodingError) as e:
            if not self.retry_conxn_errors:
                raise
            logger.warning('Connection error retrying %s', e)
            err = e
        # Comes here only if there was either exception or login was not
        # successful
        if self.retry_wait_time:
            time.sleep(self.retry_wait_time)
        self.num_session_retries += 1
        if self.num_session_retries > self.max_session_retries:
            self.num_session_retries = 0
            logger.error("Giving up after %d retries connection failure %s" % (
                self.max_session_retries, True))
            raise err
        self.authenticate_session()
        return



import argparse
import json

from avi.sdk.avi_api import ApiSession
from avi.sdk.utils.api_utils import ApiUtils
from requests.packages import urllib3
urllib3.disable_warnings()


HELP_STR = '''
Example of fetching vs metrics for a given se
    python se_vs_metrics.py -c 10.10.25.42 -p xxxx --limit 1 --se_uuid 'se-005056b01755'
'''

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=(HELP_STR))
    parser.add_argument('-t', '--tenant', help='tenant name',
                        default=None)
    parser.add_argument('--se_uuid', help='se uuid')
    parser.add_argument('--vs_metrics', help='VS Metrics to fetch for SE',
                        default='l4_client.avg_bandwidth')

    parser.add_argument('--per_vs_metrics',
                        help='Provide per vs or Aggregated metrics',
                        action='store_true')
    parser.add_argument('--step',
                        help='Granularity of metrics. Eg. 5, 300, 3600',
                        type=int, default=300)
    parser.add_argument('--limit',
                        help='number of values',
                        type=int, default=2)
    parser.add_argument('-c', '--controller',
                        help='controller ip', default='127.0.0.1')
    parser.add_argument('-u', '--username',
                        help='user name', default='admin')
    parser.add_argument('-p', '--password',
                        help='password', default='abc')

    args = parser.parse_args()
    print('parsed args', args)

    mq = {
        'metric_id': args.vs_metrics,
        'serviceengine_uuid': args.se_uuid,
        'tenant': (args.tenant if args.tenant else 'admin'),
        'step': args.step,
        'limit': args.limit,
        'aggregate_entity': not args.per_vs_metrics,
        'entity_uuid': '*',
        'pad_missing_data': False
    }

    api_ssn = ApiSession(args.controller, args.username, args.password,
                         args.tenant)
    api_utils = ApiUtils(api_ssn)

    rsp = api_utils.get_metrics_collection(tenant=args.tenant,
                                           metric_requests=[mq])
    print 'metrics query', mq
    print json.dumps(rsp, indent=2)


#!/usr/bin/python
import os
import json
from avi.sdk.avi_api import ApiSession

def DisableVirtualService(vs_name):
    """
    Creates a session using token and other user information passed to the control script.
    :param vs_name: Name of the Virtual Service which need to be disabled. 
    """
    # Get session on the basis of authentication token
    token=os.environ.get('API_TOKEN')
    user=os.environ.get('USER')
    tenant=os.environ.get('TENANT')
    with ApiSession("localhost", user, token=token, tenant=tenant) as session:
        # Get the virtualservice objct of name myVirtualService
        vs_obj = session.get_object_by_name('virtualservice', vs_name)

        if vs_obj:
            # Update the object
            vs_obj['enabled'] = False

            # Save the object
            resp = session.put('virtualservice/%s' %vs_obj['uuid'], data=vs_obj)
            print(resp.status_code, resp.json())

if __name__ == "__main__":
    DisableVirtualService('myVirtualService')
